// clang-format off

/*

 * matt | MatterLinux package manager 
 * MatterLinux 2023-2024 (https://matterlinux.xyz)

 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.

 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <https://www.gnu.org/licenses/>.

*/

// clang-format on

#include <libmp/all.h>
#include <libmp/ctx.h>
#include <libmp/error.h>
#include <libmp/util.h>
#include <locale.h>
#include <signal.h>
#include <stdio.h>
#include <stdlib.h>

#include "args.h"
#include "cmd.h"
#include "config.h"
#include "intl.h"
#include "lock.h"
#include "log.h"
#include "util.h"

lm_ctx_t ctx;

void handler(int sig) {
  unlock(&ctx);
  printf("\e[?25h\n");
  if (SIGSEGV == sig)
    error(_("Received a segfault"));
  exit(1);
}

int main(int argc, char *argv[], char *envp[]) {
  signal(SIGSEGV, handler);
  signal(SIGINT, handler);

  setbuf(stdout, NULL);
  setlocale(LC_ALL, "");
  textdomain("matt");

  cmd_t commands[] = {
      {.name = "info",    .desc = "show information about a single package", .func = cmd_info   },
      {.name = "list",    .desc = "list all the installed packages",         .func = cmd_list   },
      {.name = "sync",    .desc = "update the pool info and package lists",  .func = cmd_sync   },
      {.name = "install", .desc = "install package(s) from remote pools",    .func = cmd_install},
      {.name = "remove",  .desc = "remove installed package(s)",             .func = cmd_remove },
      {.name = "update",  .desc = "update installed package(s)",             .func = cmd_update },
  };

  char    *config_file = NULL, *root_dir = NULL;
  char    *full_datadir = NULL, *full_tmpdir = NULL;
  args_t  *args = NULL;
  bool     ret  = false;
  config_t config;

  set_env(envp);
  lm_ctx_init(&ctx);
  args = args_parse(argc, argv);

  if ((config_file = args_get_string(args, "config")) == NULL)
    config_file = "/etc/matt/config.ini";

  if (!config_load(&ctx, &config, config_file))
    goto end;

  if (args->count <= 0 || args->list[0].name != NULL)
    goto help;

  if ((root_dir = args_get_string(args, "root")) == NULL)
    root_dir = "/";

  if (!lm_ctx_set_root(&ctx, root_dir)) {
    if (LM_ERR_CtxRootNoWrite == lm_error())
      error(_("Failed to access the root directory, are you running as root?"));
    else
      error(_("Bad root directory (%s): %s"), root_dir, lm_strerror());
    goto end;
  }

  full_datadir = join_alloc(root_dir, config.datadir);
  full_tmpdir  = join_alloc(root_dir, config.tmpdir);

  if (!lm_ctx_set_data(&ctx, full_datadir)) {
    error(_("Bad datadir (%s): %s"), full_datadir, lm_strerror());
    goto end;
  }

  if (!lm_ctx_set_temp(&ctx, full_tmpdir)) {
    error(_("Bad tmpdir (%s): %s"), full_tmpdir, lm_strerror());
    goto end;
  }

  pool_config_t *pool = config.pools;
  while (NULL != pool) {
    if (!lm_ctx_pool_add(&ctx, pool->name, pool->url)) {
      error(_("Failed to add pool " FG_BOLD "%s" FG_RESET " to the list: %s"), pool->name, lm_strerror());
      goto end;
    }
    pool = pool->next;
  }

  for (int i = 0; i < sizeof(commands) / sizeof(cmd_t); i++) {
    if (!eq(commands[i].name, args->list[0].value))
      continue;

    switch (lock(&ctx)) {
    case ALREADY_LOCKED:
      error(_("Failed to lock, matt is already running"));
      goto end;

    case LOCK_ERROR:
      error(_("Failed to lock, are you root?"));
      goto end;

    default:
      break;
    }

    ret = commands[i].func(&ctx, &config, args);
    goto end;
  }

  error(_("Command not found: " FG_BOLD "%s"), args->list[0].value);
help:
  info(_("MatterLinux package manager (version %s)"), VERSION);
  info(_("Usage: " FG_BOLD "%s [command] <options> <arguments>"), argv[0]);
  printf("\n");
  info(_("Here is a list of available commands:"));

  for (int i = 0; i < sizeof(commands) / sizeof(cmd_t); i++)
    printf("    " FG_BOLD "%s" FG_RESET ":\t %s\n", commands[i].name, commands[i].desc);
  printf("\n");

  info(_("To list different options, use commands with " FG_BOLD "--help" FG_RESET " option"));
  info(_("Here is a list of available global options:"));

  printf(
      _("    " FG_BOLD "--config" FG_RESET ":\t specify the configuration file (default is /etc/matt/config.ini)\n"));
  printf(_("    " FG_BOLD "--root" FG_RESET ":\t specify a custom root directory (default is /)\n\n"));

  info(_("Licensed under GPLv3, see https://www.gnu.org/licenses/ for more information"));

  ret = true;
  goto end;

end:
  unlock(&ctx);

  free(full_datadir);
  free(full_tmpdir);

  args_free(args);
  lm_ctx_free(&ctx);
  config_free(&config);

  return ret ? EXIT_SUCCESS : EXIT_FAILURE;
}
